"""
monobit.formats.mac.dfont - MacOS suitcases and resources

(c) 2019--2023 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

import logging
from collections import Counter
from itertools import accumulate
from io import BytesIO

from ...struct import big_endian as be
from ...magic import FileFormatError
from ...streams import Stream
from ...properties import Props, reverse_dict
from ...pack import Pack

from ..sfnt import load_sfnt, save_sfnt, MAC_ENCODING, STYLE_MAP
from .nfnt import (
    extract_nfnt, convert_nfnt,
    subset_for_nfnt, convert_to_nfnt, nfnt_data_to_bytes, generate_nfnt_header
)
from .fond import extract_fond, convert_fond, create_fond


##############################################################################
# encoding constants

# 'uninterpreted symbols'
_FALLBACK_ENCODING = 32

# fonts which claim mac-roman encoding but aren't
NON_ROMAN_NAMES = {
    # https://www.unicode.org/Public/MAPPINGS/VENDORS/APPLE/SYMBOL.TXT
    # > The Mac OS Symbol encoding shares the script code smRoman
    # > (0) with the Mac OS Roman encoding. To determine if the Symbol
    # > encoding is being used, you must check if the font name is
    # > "Symbol".
    'Symbol': 'mac-symbol',
    'Cairo': '',
    'Taliesin': '',
    'Mobile': '',
    'Zapf Dingbats': '',
}


# Apple IIgs Technote #41, Font Family Numbers
_FONT_NAMES = {
    # 0: 'System Font',
    # 1: 'System Font',
    2: 'New York',
    3: 'Geneva',
    4: 'Monaco',
    5: 'Venice',
    6: 'London',
    7: 'Athens',
    8: 'San Francisco',
    9: 'Toronto',
    11: 'Cairo',
    12: 'Los Angeles',
    13: 'Zapf Dingbats',
    14: 'Bookman',
    15: 'Helvetica Narrow',
    16: 'Palatino',
    18: 'Zapf Chancery',
    20: 'Times',
    21: 'Helvetica',
    22: 'Courier',
    23: 'Symbol',
    24: 'Taliesin',
    33: 'Avant Garde',
    34: 'New Century Schoolbook',
    65533: 'Chicago',
    65534: 'Shaston',
}


##############################################################################
# resource fork/dfont format
# see https://developer.apple.com/library/archive/documentation/mac/pdf/MoreMacintoshToolbox.pdf

# Page 1-122 Figure 1-12 Format of a resource header in a resource fork
_RSRC_HEADER = be.Struct(
    data_offset='uint32',
    map_offset='uint32',
    data_length='uint32',
    map_length='uint32',
    # header is padded with zeros to 256 bytes
    # https://github.com/fontforge/fontforge/blob/master/fontforge/macbinary.c
    reserved='240s',
)

# Figure 1-13 Format of resource data for a single resource
_DATA_HEADER = be.Struct(
    length='uint32',
    # followed by `length` bytes of data
)

# Figure 1-14 Format of the resource map in a resource fork
_MAP_HEADER = be.Struct(
    reserved_header='16s',
    reserved_handle='4s',
    reserved_fileref='2s',
    attributes='uint16',
    type_list_offset='uint16',
    name_list_offset='uint16',
    # number of types minus 1
    last_type='uint16',
    # followed by:
    # type list
    # reference lists
    # name list
)
# Figure 1-15 Format of an item in a resource type list
_TYPE_ENTRY = be.Struct(
    rsrc_type='4s',
    # number of resources minus 1
    last_rsrc='uint16',
    ref_list_offset='uint16',
)

# Figure 1-16 Format of an entry in the reference list for a resource type
_REF_ENTRY = be.Struct(
    rsrc_id='uint16',
    name_offset='uint16',
    attributes='uint8',
    # we need a 3-byte offset, will have to construct ourselves...
    data_offset_hi='uint8',
    data_offset='uint16',
    reserved_handle='4s',
)

# Figure 1-17 Format of an item in a resource name list
# 1-byte length followed by bytes


def parse_resource_fork(data, formatstr=''):
    """Parse a bare resource and convert to fonts."""
    resource_table = _extract_resource_fork_header(data)
    rsrc = _extract_resources(data, resource_table)
    logging.debug(rsrc)
    directory = _construct_directory(rsrc)
    fonts = _convert_mac_font(rsrc, directory, formatstr)
    return fonts


def _extract_resource_fork_header(data):
    """Read a Classic MacOS resource fork header."""
    rsrc_header = _RSRC_HEADER.from_bytes(data)
    map_header = _MAP_HEADER.from_bytes(data, rsrc_header.map_offset)
    type_array = _TYPE_ENTRY.array(map_header.last_type + 1)
    # +2 because the length field is considered part of the type list
    type_list_offset = rsrc_header.map_offset + map_header.type_list_offset + 2
    type_list = type_array.from_bytes(data, type_list_offset)
    resources = []
    for type_entry in type_list:
        ref_array = _REF_ENTRY.array(type_entry.last_rsrc + 1)
        ref_list = ref_array.from_bytes(
            data, type_list_offset -2 + type_entry.ref_list_offset
        )
        for ref_entry in ref_list:
            # get name from name list
            if ref_entry.name_offset == 0xffff:
                name = ''
            else:
                name_offset = (
                    rsrc_header.map_offset + map_header.name_list_offset
                    + ref_entry.name_offset
                )
                name_length = data[name_offset]
                # should be ascii, but use mac-roman just in case
                name = data[name_offset+1:name_offset+name_length+1].decode('mac-roman')
            # construct the 3-byte integer
            data_offset = ref_entry.data_offset_hi * 0x10000 + ref_entry.data_offset
            offset = rsrc_header.data_offset + _DATA_HEADER.size + data_offset
            resources.append((type_entry.rsrc_type, ref_entry.rsrc_id, offset, name))
    return resources


def _extract_resources(data, resources):
    """Extract resources."""
    parsed_rsrc = []
    for rsrc_type, rsrc_id, offset, name in resources:
        if rsrc_type == b'FOND':
            logging.debug(
                'Font family resource #%d: type FOND name `%s`', rsrc_id, name
            )
            parsed_rsrc.append((
                rsrc_type, rsrc_id, dict(
                    name=name, **extract_fond(data, offset)
                )
            ))
        elif rsrc_type == b'FONT' and name and not (rsrc_id % 128):
            # rsrc_id % 128 is the point size
            logging.debug(
                'Name entry #%d: type FONT name `%s`',
                rsrc_id, name
            )
            # inside macintosh:
            # > Since 0 is not a valid font size, the resource ID having
            # > 0 in the size field is used to provide only the name of
            # > the font: The name of the resource is the font name. For
            # > example, for a font named Griffin and numbered 200, the
            # > resource naming the font would have a resource ID of 25600
            # > and the resource name 'Griffin'. Size 10 of that font would
            # > be stored in a resource numbered 25610.
            # keep the name in the directory table
            parsed_rsrc.append((
                b'', rsrc_id, dict(name=name),
            ))
        elif rsrc_type in (b'NFNT', b'FONT'):
            logging.debug(
                'Bitmapped font resource #%d: type %s name `%s`',
                rsrc_id, rsrc_type.decode('mac-roman'), name
            )
            parsed_rsrc.append((
                rsrc_type, rsrc_id, extract_nfnt(data, offset)
            ))
        elif rsrc_type == b'sfnt':
            logging.debug(
                'TrueType font resource #%d: type %s name `%s`',
                rsrc_id, rsrc_type.decode('mac-roman'), name
            )
            bytesio = Stream.from_data(data[offset:], mode='r')
            fonts = load_sfnt(bytesio)
            parsed_rsrc.append((
                rsrc_type, rsrc_id, dict(fonts=fonts)
            ))
        else:
            logging.debug(
                'Skipped resource #%d: type %s name `%s`',
                rsrc_id, rsrc_type.decode('mac-roman'), name
            )
    return parsed_rsrc


def _construct_directory(parsed_rsrc):
    """Construct font family directory."""
    info = {}
    for rsrc_type, rsrc_id, kwargs in parsed_rsrc:
        # new-style directory entries
        if rsrc_type == b'FOND':
            props = convert_fond(**kwargs)
            info.update(props)
        # old-style name-only FONT resources (font_size 0)
        elif rsrc_type == b'':
            font_number = rsrc_id // 128
            info[font_number] = {'family': kwargs['name']}
    return info


def _convert_mac_font(parsed_rsrc, info, formatstr):
    """convert properties and glyphs."""
    fonts = []
    for rsrc_type, rsrc_id, kwargs in parsed_rsrc:
        if rsrc_type == b'sfnt':
            rsrc_fonts = kwargs['fonts']
            rsrc_fonts = (
                _font.modify(
                    source_format = f'[Mac] {_font.source_format}',
                )
                for _font in rsrc_fonts
            )
            fonts.extend(rsrc_fonts)
        elif rsrc_type in (b'FONT', b'NFNT'):
            format = ''.join((
                rsrc_type.decode('mac-roman'),
                f' in {formatstr}' if formatstr else ''
            ))
            props = {
                'family': kwargs.get('name', '') or f'{rsrc_id}',
                'source_format': f'[Mac] {format}',
            }
            if rsrc_type == b'FONT':
                # get name and size info from resource ID
                # https://developer.apple.com/library/archive/documentation/mac/Text/Text-191.html#HEADING191-0
                # > The resource ID of the font must equal the number produced by
                # > concatenating the font ID times 128 with the font size.
                # > Remember that fonts stored in 'FONT' resources are restricted
                # > to a point size of less than 128 and to a font ID in the range
                # > 0 to 255. The resource ID is computed by the following formula:
                # >     resourceID := (font ID * 128) + font size;
                font_number, font_size = divmod(rsrc_id, 128)
                # we've already filtered out the case font_size == 0
                props.update({
                    'point_size': font_size,
                    'family': _FONT_NAMES.get(font_number, str(font_number))
                })
                # prefer directory info to info inferred from resource ID
                # (in so far provided by FOND or directory FONT)
                props.update(info.get(rsrc_id, info.get(font_number, {})))
            else:
                # update properties with directory info
                props.update(info.get(rsrc_id, {}))
            if 'encoding' not in props or props.get('family', '') in NON_ROMAN_NAMES:
                props['encoding'] = NON_ROMAN_NAMES.get(props.get('family', ''), 'mac-roman')
            font = convert_nfnt(props, **kwargs)
            if font.glyphs:
                font = font.label()
                fonts.append(font)
    return fonts


###############################################################################
# dfont writer

# from FontForge notes https://fontforge.org/docs/techref/macformats.html :
# > When an ‘sfnt’ resource contains a font with a multibyte encoding (CJK or
# > unicode) then the ‘FOND’ does not have entries for all the characters. The
# > ‘sfnt’ will (always?) have a MacRoman encoding as well as the multibyte encoding
# > and the ‘FOND’ will contain information on just that subset of the font. (I have
# > determined this empirically, I have seen no documentation on the subject)
#
# > Currently bitmap fonts for multibyte encodings are stored inside an sfnt
# > (truetype) resource in the ‘bloc’ and ‘bdat’ tables. When this happens there are
# > a series of dummy ‘NFNT’ resources in the resource file, one for each strike.
# > Each resource is 26 bytes long (which means they contain the FontRec structure
# > but no data tables) and are flagged by having rowWords set to 0. (I have
# > determined this empirically, I have seen no documentation on the subject)


def save_dfont(fonts, outstream, resource_type):
    """
    Save font to MacOS resource fork or data-fork resource.

    resource_type: type of resource to store font in. One of `sfnt`, `NFNT`.
    """
    resource_type = resource_type.lower()
    if resource_type not in ('sfnt', 'nfnt'):
        raise ValueError(
            'Only saving to sfnt or NFNT resource currently supported'
        )
    resources = []
    if resource_type == 'sfnt':
        sfnt_io = BytesIO()
        result = save_sfnt(fonts, sfnt_io)
        font, *_ = fonts
        family_id = _get_family_id(font.family, font.encoding)
        resources.append(
            Props(type=b'sfnt', id=family_id, name='', data=sfnt_io.getvalue()),
        )
    # reduce fonts to what's storable in (stub) FOND/NFNT
    # we need a Pack for _group_families
    fonts = Pack(subset_for_nfnt(_f) for _f in fonts)
    for family_id, style_group in _group_families(fonts):
        i = 0
        for style_id, size_group in style_group:
            for font in size_group:
                if resource_type == 'sfnt':
                    # create stub NFNT if the bitmaps are in an sfnt
                    nfnt_data = generate_nfnt_header(font, endian='big')
                else:
                    nfnt_data = convert_to_nfnt(
                        font, endian='big', ndescent_is_high=True,
                        create_width_table=True, create_height_table=False,
                    )
                resources.append(
                    Props(
                        type=b'NFNT',
                        # note that we calculate this *separately* in the FOND builder
                        id=family_id + i,
                        # are there any specifications for the name?
                        name=font.name,
                        data=nfnt_data_to_bytes(nfnt_data),
                    ),
                )
                i += 1
        fond_data = create_fond(style_group, family_id)
        resources.append(
            Props(
                type=b'FOND', id=family_id,
                name=font.family, data=fond_data,
            ),
        )
    _write_resource_fork(outstream, resources)


def _get_family_id(name, encoding):
    """Generate a resource id based on the font's properties."""
    script_code = reverse_dict(MAC_ENCODING).get(encoding, _FALLBACK_ENCODING)
    return _hash_to_id(name, script=script_code)


def _write_resource_fork(outstream, resources):
    """
    Write a Mac dfont/resource fork.

    resources: list of ns(type, id, name, data)
    """
    # order resources by type (so all resources of the same type are consecutive
    resources.sort(key=lambda _res: _res.type)
    # counter follows insertion order
    types = Counter(_res.type for _res in resources)
    # construct the resource map
    map_header = _MAP_HEADER(
        # not sure what these are
        attributes=0,
        # type list comes straight after this header
        # -2 because the last_type in this header definition
        # is counted as part of the type list
        type_list_offset=_MAP_HEADER.size-2, #28,
        # after type list plus reference lists
        name_list_offset=(
            _MAP_HEADER.size
            + _TYPE_ENTRY.size * len(types)
            + _REF_ENTRY.size * len(resources)
        ),
        # number of types minus 1
        last_type=len(types) - 1,
    )
    # construct the type list
    type_list = [
        _TYPE_ENTRY(
            rsrc_type=_type,
            last_rsrc=_count - 1,
            # ref_list_offset='uint16',
        )
        for _type, _count in types.items()
    ]
    offset = 2 + _TYPE_ENTRY.size * len(type_list)
    for entry in type_list:
        entry.ref_list_offset = offset
        offset += _REF_ENTRY.size * (entry.last_rsrc + 1)
    type_list = (_TYPE_ENTRY * len(types))(*type_list)
    # construct the name list
    name_list = tuple(
        bytes((len(_res.name),)) + _res.name.encode('mac-roman')
        for _res in resources
    )
    name_offsets = accumulate((len(_n) for _n in name_list), initial=0)
    data_offsets = accumulate(
        # include 4 bytes for length field
        (4 + len(_res.data) for _res in resources),
        initial=0
    )
    # construct the reference list
    reference_list = (
        _REF_ENTRY(
            rsrc_id=_res.id,
            name_offset=-1 if not _res.name else _name_offset,
            attributes=0,
            # we need a 3-byte offset, will have to construct ourselves...
            data_offset_hi=_data_offset // 0x1000,
            data_offset=_data_offset % 0x1000,
        )
        for _res, _name_offset, _data_offset in zip(resources, name_offsets, data_offsets)
    )
    reference_list = (_REF_ENTRY * len(resources))(*reference_list)
    rsrc_map = (
        bytes(map_header)
        + bytes(type_list)
        + bytes(reference_list)
        + b''.join(name_list)
    )
    data = b''.join(
        bytes(be.uint32(len(_res.data))) + _res.data
        for _res in resources
    )
    rsrc_header = _RSRC_HEADER(
        # data come right after the header and padding
        data_offset=256,
        map_offset=256 + len(data),
        data_length=len(data),
        map_length=len(rsrc_map),
    )
    outstream.write(
        bytes(rsrc_header)
        + data
        + rsrc_map
    )


# family name hash algorithm
# ported from https://github.com/zoltan-dulac/fondu/blob/master/ufond.c
#
# FONDU licence:
# > PfaEdit is copyright (C) 2000,2001,2002,2003 by George Williams
# >
# >    Redistribution and use in source and binary forms, with or without
# >    modification, are permitted provided that the following conditions are met:
# >
# >    Redistributions of source code must retain the above copyright notice, this
# >    list of conditions and the following disclaimer.
# >
# >    Redistributions in binary form must reproduce the above copyright notice,
# >    this list of conditions and the following disclaimer in the documentation
# >    and/or other materials provided with the distribution.
# >
# >    The name of the author may not be used to endorse or promote products
# >    derived from this software without specific prior written permission.
# >
# >    THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
# >    WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# >    MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
# >    EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# >    SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# >    PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# >    OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# >    WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# >    OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# >    ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# >
# > The configure script is subject to the GNU public license. See the file
# > COPYING.

def _hash_to_id(family_name, script):
    """Generate a resource id based on the font family name."""
    low = 128
    high = 0x4000
    hash = 0
    if script:
        low = 0x4000 + (script-1)*0x200;
        high = low + 0x200;
    for ch in family_name:
        temp = (hash>>28) & 0xf
        hash = (hash<<4) | temp
        hash ^= ord(ch) - 0x20
    hash %= (high-low)
    hash += low
    return hash


def mac_style_from_name(style_name):
    """Get font style from human-readable representation."""
    return sum((2<<_bit for _bit, _k in STYLE_MAP.items() if _k in style_name))


def _group_families(fonts):
    """Group pack of fonts by families and subfamilies."""
    families = tuple(sorted(
        # assuming encoding stays the same across family
        (_get_family_id(_name, _group[0].encoding), _group)
        for _name, _group in fonts.itergroups('family')
    ))
    fond_families = []
    for id, group in families:
        chars = set(_font.get_codepoints() for _font in group)
        if len(set(chars)) > 1:
            logging.warning(
                "Can't combine fonts into families: different character ranges."
            )
            return _group_individually(fonts)
        spacings = set(_font.spacing for _font in group)
        if len(set(spacings)) > 1:
            logging.warning(
                "Can't combine fonts into families: different spacing characteristics."
            )
            return _group_individually(fonts)
        style_groups = tuple(sorted(
            (mac_style_from_name(_subfamily), _fonts)
            for _subfamily, _fonts in group.itergroups('subfamily')
        ))
        for _, group in style_groups:
            sizes = tuple(_f.point_size for _f in group)
            if len(sizes) != len(set(sizes)):
                logging.warning(
                    "Can't combine fonts into families: duplicate style and size."
                )
            return _group_individually(fonts)
        fond_families.append((id, style_groups))
    return fond_families


def _group_individually(fonts):
    """Return each individual font as its own family."""
    fond_families = tuple(sorted(
        (
            # we're assuming names differ
            _get_family_id(_font.name, _font.encoding),
            ((mac_style_from_name(_font.subfamily), [_font]),)
        )
        for _font in fonts
    ))
    return fond_families



"""
monobit.formats.alto - Xerox Alto .AL screen font format

(c) 2023 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

import logging
from itertools import count

from ..struct import bitfield, flag, big_endian as be
from ..storage import loaders, savers
from ..magic import FileFormatError
from ..properties import Props
from ..font import Font
from ..glyph import Glyph
from ..raster import Raster
from ..labels import Tag


@loaders.register(
    name='alto',
    patterns=('*.al',),
)
def load_alto(instream):
    """Load font from Xerox Alto screen font file."""
    header, al_glyphs =  _read_alto(instream)
    glyphs = _convert_glyphs(header, al_glyphs)
    return Font(glyphs, line_height=header.height)


@loaders.register(
    name='bitblt',
    patterns=('*.strike',),
)
def load_bitblt(instream):
    """Load font from Xerox Alto .strike file."""
    props, glyphs =  _read_bitblt(instream)
    return Font(glyphs, **props)


@loaders.register(
    name='prepress',
    patterns=('*.ac',),
)
def load_prepress(instream):
    """Load font from Xerox Alto PrePress .ac file."""
    props, glyphs =  _read_prepress(instream)
    return Font(glyphs, **props)


##############################################################################
# .AL "CONVERT" format
# http://www.bitsavers.org/pdf/xerox/alto/printing/AltoFontFormats_Oct1980.pdf

_AL_HEADER = be.Struct(
    height='uint16',
    proportional=bitfield('uint8', 1),
    baseline=bitfield('uint8', 7),
    maxWidth='uint8',
)

_XH_DATA = be.Struct(
    XW='uint16',
    HD='uint8',
    XH='uint8',
)

def _read_alto(instream):
    """Read Xerox Alto screen font file."""
    header = _AL_HEADER.read_from(instream)
    # character pointer table
    # pointers are self-referenced word offsets to the XW value in character structs
    pointers = []
    i = 0
    while True:
        here = instream.tell()
        # beak if we are at the first pointed-to location
        if pointers and here >= min(pointers):
            break
        pointer = be.uint16.read_from(instream)
        # break on null pointer
        if not int(pointer):
            break
        # record absolute position in file
        pointers.append(here + pointer*2)
    al_glyphs = tuple(
        _read_glyph(instream, header, pointer, cp)
        for cp, pointer in enumerate(pointers)
    )
    return header, al_glyphs


def _read_glyph(instream, header, pointer, cp):
    instream.seek(pointer)
    xh_data = _XH_DATA.read_from(instream)
    instream.seek(-_XH_DATA.size - xh_data.XH*2, 1)
    glyph_data = instream.read(xh_data.XH*2)
    return Props(
        pixels=Raster.from_bytes(glyph_data, width=16),
        **vars(xh_data)
    )


def _convert_glyphs(header, al_glyphs):
    glyphs = tuple(
        _convert_glyph(header, props)
        for props in al_glyphs
    )
    final_glyphs = []
    for cp, g in enumerate(glyphs[:256]):
        if g:
            while hasattr(g, 'extension'):
                ext = glyphs[g.extension]
                if g.shift_up > ext.shift_up:
                    g = g.expand(bottom=g.shift_up - ext.shift_up)
                else:
                    ext = ext.expand(bottom=ext.shift_up - g.shift_up)
                if g.height > ext.height:
                    ext = ext.expand(top=g.height - ext.height)
                else:
                    g = g.expand(top=ext.height - g.height)
                g = Glyph(
                    Raster.concatenate(g.pixels, ext.pixels),
                    right_bearing=ext.right_bearing, shift_up=ext.shift_up
                )
            final_glyphs.append(g.modify(codepoint=cp))
    return final_glyphs


def _convert_glyph(header, props):
    if (props.XW % 2):
        advance_width = (props.XW-1) // 2
        extension = None
    else:
        advance_width=16
        # index of extension char
        extension = props.XW // 2
    glyph_props = dict(
        right_bearing=-16 + advance_width,
        shift_up=header.baseline-props.HD-props.XH+1,
        extension=extension,
    )
    if not props.pixels:
        if props.HD or props.XW > 1:
            return Glyph.blank(width=16, height=props.HD, **glyph_props)
    else:
        return Glyph(props.pixels, width=16, **glyph_props)
    return None


##############################################################################
# "plainstrike" or "BITBLT" format

# a variant with a separate StrikeIndex file (.StrikeX) exists, but:
# > to the best of my knowledge, no one has ever used a StrikeIndex format

_FORMAT_STRUCT = be.Struct(
    # always =1, meaning 'new style'
    oneBit=bitfield('uint16', 1),
    # =1 means StrikeIndex, =0 otherwise
    index=bitfield('uint16', 1),
    # =1 if all characters have same value of Wx, else =0
    fixed=bitfield('uint16', 1),
    # =1 if KernedStrike, =0 if PlainStrike
    kerned=bitfield('uint16', 1),
    blank=bitfield('uint16', 12),
)

_STRIKE_HEADER = be.Struct(
    format=_FORMAT_STRUCT,
    # minimum character code
    min='uint16',
    # maximum character code
    max='uint16',
    # maximum spacing width of any character = max{Wx}
    maxwidth='uint16',
)

_STRIKE_BODY = be.Struct(
    # number of words in the StrikeBody
    length='uint16',
    # number of scan-lines above the baseline
    ascent='uint16',
    # number of scan-lines below the baseline
    descent='uint16',
    # always =0 [used to be used for padding schemes]
    xoffset='uint16',
    # number of words per scan-line in the strike
    raster='uint16',
    # the bit map, where height = ascent + descent
    # bitmap word raster*height
    # pointers into the strike, indexed by code
    # xinsegment ↑ min, max+2 word
)

_BOUNDING_BOX_BLOCK = be.Struct(
    # X offset
    FBBox='uint16',
    # Y offset
    FBBoy='uint16',
    # X width
    FBBdx='uint16',
    # Y width
    FBBdy='uint16',
)

_WIDTH_ENTRY = be.Struct(
    # the entire spacing word will be = (-1) (both bytes =377b)
    # to flag a non-existent character, else the bytes are:
    offset='int8',
    width='int8',
)

def _read_bitblt(instream):
    # @StrikeHeader
    header = _STRIKE_HEADER.read_from(instream)
    if not header.format.oneBit:
        raise FileFormatError('Not a Xerox BITBLT strike')
    if header.format.index:
        raise FileFormatError('StrikeIndex format not supported')
    # @BoundingBoxBlock
    if header.format.kerned:
        bbox = _BOUNDING_BOX_BLOCK.read_from(instream)
    # @StrikeBody
    body = _STRIKE_BODY.read_from(instream)
    height = body.ascent + body.descent
    strikebytes = instream.read(2*body.raster*height)
    # max is the highest included code; max+1 holds the replacement char
    # (if min==max we have 2 glyphs and 3 bounds)
    offsets = (be.uint16 * (header.max+3 - header.min)).read_from(instream)
    # @WidthBody
    if header.format.kerned:
        spacing = (_WIDTH_ENTRY * (header.max+2 - header.min)).read_from(instream)
    strike = Raster.from_bytes(strikebytes, 16*body.raster, height)
    # conversion section
    # convert strike to glyphs
    if header.format.kerned:
        glyphs = [
            Glyph(
                strike.crop(left=_offset, right=max(0, 16*body.raster - _next)),
                codepoint=_cp,
                shift_up=-body.descent,
                left_bearing=_spacing.offset+bbox.FBBox,
                right_bearing=(
                    _spacing.width
                    -(_next-_offset)
                    -(_spacing.offset+bbox.FBBox)
                ),
            )
            for _offset, _next, _spacing, _cp in zip(
                offsets, offsets[1:], spacing, count(header.min)
            )
            if _spacing.width >= 0 and _spacing.offset >= 0
        ]
    else:
        glyphs = [
            Glyph(
                strike.crop(left=_offset, right=max(0, 16*body.raster - _next)),
                codepoint=_cp,
                shift_up=-body.descent,
            )
            for _offset, _next, _cp in zip(offsets, offsets[1:], count(header.min))
            if _offset != _next
        ]
    # last char is replacement char (referred to as "dummy character" in the docs)
    glyphs[-1] = glyphs[-1].modify(codepoint=None, tag='dummy')
    props = dict(
        ascent=body.ascent,
        descent=body.descent,
        default_char=Tag('dummy'),
    )
    return props, glyphs


##############################################################################
# PrePress format

_IX = be.Struct(
    # Various type codes are assigned
    type=bitfield('uint16', 4),
    # Length of entry in words, counting this one
    length=bitfield('uint16', 12),
)

# name-definition index entry
_IXN = be.Struct(
    # Header with type == 1
    ix=_IX,
    # The numeric code
    code='uint16',
    # The number of characters in the name
    nameLength='uint8',
    # Room for the name
    characters='19s',
)

_CHARACTER_INDEX_ENTRY = be.Struct(
    # type = 3 for character index
    ix=_IX,
    # Family name, using a name code
    family='uint8',
    # (if bold then 2 elseif light then 4 else 0) +
    # (if italic then 1 else 0) +
    # (if codensed then 6 elseif expanded then 12 else 0) +
    # (if Xerox then 0 elseif ASCII then 18 else 36)
    face='uint8',
    # Code for the "beginning character"
    bc='uint8',
    # Code for the "ending character"
    ec='uint8',
    # Size of the font segment [in micas == 10 micron units or 1/2540 in]
    size='uint16',
    # Rotation of the font segment
    rotation='uint16',
    # Starting address in file of the font segment
    segmentSA='uint32',
    # Length of the segment
    segmentLength='uint32',
    # Resolution in scan-lines/inch * 10
    resolutionX='uint16',
    # Resolution in bits/inch * 10
    resolutionY='uint16',
)


_CHARACTER_DATA = be.Struct(
    # X Width (scan-lines)
    # fixed-point, divide by 0x10000
    Wx='uint32',
    # Y width (bits)
    # fixed-point, divide by 0x10000
    Wy='uint32',
    # Bounding box offsets
    BBox='int16',
    BBoy='int16',
    # Width of bounding box (scan-lines)
    BBdx='uint16',
    # Height of bounding box (bits) or -1 for not defined
    BBdy='int16',
)

# relFilePos=uint32

_RASTER_DEFN = be.Struct(
    # Height of raster (in words)
    BBdyW=bitfield('uint16', 6),
    # Same as BBdx in CharacterData
    BBdx=bitfield('uint16', 10),
    # followed by raster, BBdyW*BBdx words
)


def _read_prepress(instream):
    """Read a PrePress file."""
    ixn = _IXN.read_from(instream)
    if ixn.ix.type != 1:
        raise FileFormatError('Not an .ac file: no name index entry')
    cie = _CHARACTER_INDEX_ENTRY.read_from(instream)
    if cie.ix.type != 3:
        raise FileFormatError('Not an .ac file: no character index entry')
    if cie.rotation != 0:
        raise FileFormatError('Nonzero rotation not supported for this format.')
    instream.seek(cie.segmentSA*2)
    nchars = cie.ec - cie.bc + 1
    char_data = (_CHARACTER_DATA * nchars).read_from(instream)
    anchor = instream.tell()
    directory = (be.uint32 * nchars).read_from(instream)
    glyphs = []
    for cp, offset, cd in zip(range(cie.bc, cie.ec+1), directory, char_data):
        if cd.BBdy < 0:
            continue
        instream.seek(anchor+offset*2)
        raster_defn = _RASTER_DEFN.read_from(instream)
        raster = Raster.from_bytes(
            instream.read(2*raster_defn.BBdyW*raster_defn.BBdx),
            width=raster_defn.BBdyW*16,
            height=raster_defn.BBdx,
        ).turn(-1)
        glyphs.append(Glyph(
            raster, codepoint=cp,
            left_bearing=cd.BBox,
            shift_up=cd.BBoy,
            right_bearing=cd.Wx//0x10000-cd.BBox-cd.BBdx,
            # ignoring Wy - vertical advance
        ))
    # decode face number
    encoding = None
    setwidth = 'normal'
    face = cie.face
    if face >= 36:
        face -= 36
    elif face >= 18:
        encoding = 'ascii'
        face -= 18
    else:
        encoding = 'xerox'
    if face >= 12:
        setwidth = 'expanded'
        face -= 12
    elif face >= 6:
        setwidth = 'condensed'
        face -= 6
    weight = ('bold' if face&2 else 'light' if face&4 else 'regular')
    slant = ('italic' if face&1 else 'roman')
    # remaining properties
    props = dict(
        family=ixn.characters.decode('ascii', 'replace'),
        weight=weight,
        slant=slant,
        setwidth=setwidth,
        # encoding flag isn't set right in any of the .ac files seen
        #encoding=encoding,
        dpi=(cie.resolutionX//10, cie.resolutionY//10),
    )
    return props, glyphs



"""
monobit.formats.prebuilt - Adobe BE/LE Prebuilt Format

(c) 2023 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

import logging
from itertools import cycle

from ..struct import bitfield, little_endian as le, big_endian as be
from ..properties import Props
from ..storage import loaders, savers
from ..font import Font
from ..glyph import Glyph
from ..magic import FileFormatError
from ..taggers import tagmaps


@loaders.register(
    name='prebuilt',
    patterns=('*.bepf', '*.lepf'),
)
def load_pf(instream):
    """Load font from bepf/lepf file."""
    pf_props, pf_glyphs = _read_pf(instream)
    logging.info(pf_props)
    fonts = _convert_from_pf(pf_props, pf_glyphs)
    return fonts


################################################################################
# Adobe prebuilt format
# https://github.com/johnsonjh/NeXTDPS/blob/master/fonts-21/bin/prebuiltformat.h
# https://github.com/johnsonjh/NeXTDPS/blob/master/fonts-21/bin/prebuild.c
#
# these appear to be the source of the `prebuild` command line tool
# which converts BDF files into BEPF or LEPF.
# man page: http://macosx.polarhome.com/service/man/?qf=prebuild&tf=2&of=NeXTSTEP&sf=1

# The header file includes structures for vertical metrics but the tool does not
# convert these from BDF.
# Lacking a spec, sample files or other documentation, I don't think we can
# meaningfully convert vertical metrics in these files


# we have a choice of big and little endian files, need different base classes
_BASE = {'l': le, 'b': be}


_PREBUILTFILEVERSION = 3
_FIXEDSCALE = 65536

# typedef struct _t_PrebuiltFile {
_PREBUILT_FILE = {
    _endian: _BASE[_endian].Struct(
        # /* equal to 0 iff little-endian, otherwise big-endian */
        # byteOrder='short',
        # /* file version number (equal to PREBUILTFILEVERSION) */
        version='short',
        # /* font's fid */
        identifier='long',
        # /* file position of name of font (null terminated) */
        fontName='long',
        # /* file position of name of character set (null terminated)
        # (equal to either "ISOLatin1CharacterSet" or the name of the font) */
        characterSetName='long',
        # /* number of character in character set */
        numberChar='short',
        # /* number of distinct matrices imaged */
        numberMatrix='short',
        # /* file position of the first element of an
        # array of numberChar PrebuiltWidth's */
        widths='long',
        # /* file position of first of numberChar character name's
        # (sorted lexicographically, each null terminated) */
        names='long',
        # /* total number of bytes of character name's (including nulls) */
        lengthNames='long',
        # /* file position of the first element of an
        # array of numberMatrix PrebuiltMatrix 's */
        matrices='long',
        # /* file position of the first element of an
        # array of numberChar*numberMatrix PrebuiltMask's */
        masks='long',
        # /* equal to 0 iff vertical data NOT included,
        # otherwise file position of first element of an
        # array of numberChar PrebuiltVertWidths's */
        vertWidths='long',
        # /* equal to 0 iff vertical data NOT included,
        # otherwise file position of first element of an
        # array of numberChar*numberMatrix PrebuiltVertMetrics's */
        vertMetrics='long',
        # /* total number of bytes of mask data.  Mask
        # data are positioned in the file in the
        # same order as their corresponding PrebuiltMask. */
        lengthMaskData='long',
    )
    for _endian in ('l', 'b')
}


# typedef struct _t_PrebuiltMatrix {
_PREBUILT_MATRIX = {
    _endian: _BASE[_endian].Struct(
        # /* x' = a*x + c*y + tx (fixed)*/
        a='int32',
        # /* y' = b*x + d*y + ty (fixed)*/
        b='int32',
        # /* (fixed)*/
        c='int32',
        # /* (fixed)*/
        d='int32',
        # /* (fixed)*/
        tx='int32',
        # /* (fixed)*/
        ty='int32',
        # /* depth of mask pixel */
        depth='int32',
    )
    for _endian in ('l', 'b')
}

# typedef struct _t_PrebuiltWidth {
_PREBUILT_WIDTH = {
    _endian: _BASE[_endian].Struct(
        # /* x component of horizontal vector (fixed) */
        hx='int32',
        # /* y component of horizontal vector (fixed) */
        hy='int32',
    )
    for _endian in ('l', 'b')
}


_HCOORD = {
    _endian: _BASE[_endian].Struct(
        hx='int8',
        hy='int8',
    )
    for _endian in ('l', 'b')
}


# typedef struct _t_PrebuiltMask {
_PREBUILT_MASK = {
    _endian: _BASE[_endian].Struct(
        # /* width of mask in pixels */
        width='uint8',
        # /* height of mask in pixels */
        height='uint8',
        # /* horizontal offsets */
        maskOffset=_HCOORD[_endian],
        # /* horizontal widths */
        maskWidth=_HCOORD[_endian],
        # /* pad to make multiple of 4 bytes; */
        pad='short',
        # /* file position of mask data ((width * depth + 7) / 8 * height bytes) */
        maskData='long',
    )
    for _endian in ('l', 'b')
}


# typedef struct _t_PrebuiltVertWidths {
_PREBUILT_VERT_WIDTHS = {
    _endian: _BASE[_endian].Struct(
        # /* x component of vertical vector; (fixed) */
        vx='int32',
        # /* y component of vertical vector; (fixed) */
        vy='int32',
    )
    for _endian in ('l', 'b')
}


_VCOORD = {
    _endian: _BASE[_endian].Struct(
        vx='int8',
        vy='int8',
    )
    for _endian in ('l', 'b')
}

# typedef struct _t_PrebuiltVertMetrics {
_PREBUILT_VERT_METRICS = {
    _endian: _BASE[_endian].Struct(
        # /* vertical offsets */
        offset=_VCOORD[_endian],
        # /* vertical widths */
        width=_VCOORD[_endian]
    )
    for _endian in ('l', 'b')
}



# assumed max string length
MAX_STRING = 256

def _read_pf(instream):
    """Read an Adobe prebuilt file."""
    pf_props = Props()
    pf_props.byteOrder = instream.read(2) != b'\0\0'
    endian = 'b' if pf_props.byteOrder else 'l'
    header = _PREBUILT_FILE[endian].read_from(instream)
    pf_props |= Props(**vars(header))
    instream.seek(header.fontName)
    pf_props.fontName, _, _ = instream.read(MAX_STRING).partition(b'\0')
    instream.seek(header.characterSetName)
    pf_props.characterSetName, _, _ = instream.read(MAX_STRING).partition(b'\0')
    instream.seek(header.names)
    pf_props.names = instream.read(header.lengthNames-1).split(b'\0')
    instream.seek(header.widths)
    pf_props.widths = (_PREBUILT_WIDTH[endian] * header.numberChar).read_from(instream)
    instream.seek(header.matrices)
    pf_props.matrices = (_PREBUILT_MATRIX[endian] * header.numberMatrix).read_from(instream)
    if header.vertWidths or header.vertMetrics:
        logging.warning(
            'Vertical metrics for this format not supported due to lack of documentation. '
            'Please consider sharing your input file as a format sample '
            'in an issue report at https://github.com/robhagemans/monobit/issues'
        )
    if header.vertWidths:
        instream.seek(header.vertWidths)
        pf_props.vertWidths = (_PREBUILT_VERT_WIDTHS[endian] * header.numberChar).read_from(instream)
    else:
        pf_props.vertWidths = (None,) * header.numberChar
    if header.vertMetrics:
        instream.seek(header.vertMetrics)
        pf_props.vertMetrics = ((_PREBUILT_VERT_METRICS[endian] * header.numberChar) * header.numberMatrix).read_from(instream)
    else:
        pf_props.vertMetrics = ((None,) * header.numberChar,) * header.numberMatrix
    instream.seek(header.masks)
    # glyph data in pf_props.masks
    pf_masks = ((_PREBUILT_MASK[endian] * header.numberChar) * header.numberMatrix).read_from(instream)
    pf_masks = tuple(
        tuple(Props(**vars(_mask)) for _mask in _strike)
        for _strike in pf_masks
    )
    for strike, matrix in zip(pf_masks, pf_props.matrices):
        for mask in strike:
            instream.seek(mask.maskData)
            size = ((mask.width * matrix.depth + 7) // 8) * mask.height
            mask.maskData = instream.read(size)
    return pf_props, pf_masks


def _convert_from_pf(pf_props, pf_masks):
    """Convert Adobe prebuilt format data to monobit."""
    strikes = tuple(
        tuple(
            Glyph.from_bytes(
                _m.maskData, width=_m.width, height=_m.height,
                align='left' if pf_props.byteOrder else 'right',
                tag=_name.decode('latin-1'),
                ## horizontal
                left_bearing=-_m.maskOffset.hx,
                # we don't support maskOffset.hy (cross-advance)
                right_bearing=_m.maskWidth.hx + _m.maskOffset.hx - _m.width,
                shift_up=_m.maskOffset.hy - _m.height,
                ## scalable
                # we ignore hy (scalable cross-advance)
                # this is bdf SWIDTH/1000
                # so we need to multiply by point_size * xdpi / 72 (with ydpi==75)
                # and point_size == _matrix.d / _FIXEDSCALE (below)
                scalable_width=f'{(_width.hx/_FIXEDSCALE) * (_matrix.a/_FIXEDSCALE) * (75/72):.2f}',
            )
            for _m, _name, _width in zip(_strike, pf_props.names, pf_props.widths)
        )
        for _strike, _matrix in zip(pf_masks, pf_props.matrices)
        # assume higher depth means colour or grayscale
        if _matrix.depth == 1
    )
    if len(strikes) < len(pf_masks):
        logging.warning(
            f'{len(pf_masks) - len(strikes)} bitmap strikes could not be converted: '
            'colour, grayscale and antialiased bitmaps not supported.'
        )
    # little-endian format has bit order reversed as well
    if not pf_props.byteOrder:
        strikes = tuple(
            tuple(_g.mirror(adjust_metrics=False) for _g in _strike)
            for _strike in strikes
        )
    fonts = tuple(
        Font(
            _glyphs, font_id=pf_props.identifier,
            family=pf_props.fontName.decode('latin-1'),
            # from the source code, we deduce
            #  matrix.a / fixedScale == bdf SIZE.PointSize * (75 dpi) / SIZE.Xres
            #  matrix.d / fixedScale == bdf SIZE.PointSize * (75 dpi) / SIZE.Yres
            # however the pf file doesn't include the point size or xres/yres separately
            # if we assume Yres == 75 dpi by definition,
            # d becomes point size and a gives us aspect ratio
            dpi=(75 * _matrix.a / _matrix.d, 75),
            point_size=_matrix.d / _FIXEDSCALE,
            encoding='latin-1' if pf_props.characterSetName == b'ISOLatin1CharacterSet' else '',
        ).label(char_from=tagmaps['adobe'])
        for _matrix, _glyphs in zip(pf_props.matrices, strikes)
    )
    return fonts
